/**
 * @license Copyright 2020 The Lighthouse Authors. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 */
'use strict';

/* eslint-env browser */

/** @typedef {HTMLElementTagNameMap & {[id: string]: HTMLElement}} HTMLElementByTagName */

const KiB = 1024;
const MiB = KiB * KiB;

class TreemapUtil {
  /**
   * @param {LH.Treemap.Node} node
   * @param {(node: LH.Treemap.Node, path: string[]) => void} fn
   * @param {string[]=} path
   */
  static walk(node, fn, path) {
    if (!path) path = [];
    path.push(node.name);

    fn(node, path);
    if (!node.children) return;

    for (const child of node.children) {
      TreemapUtil.walk(child, fn, [...path]);
    }
  }

  /**
   * @param {string[]} path1
   * @param {string[]} path2
   */
  static pathsAreEqual(path1, path2) {
    if (path1.length !== path2.length) return false;
    for (let i = 0; i < path1.length; i++) {
      if (path1[i] !== path2[i]) return false;
    }
    return true;
  }

  /**
   * @param {string[]} maybeSubpath
   * @param {string[]} path
   */
  static pathIsSubpath(maybeSubpath, path) {
    if (maybeSubpath.length > path.length) return false;
    for (let i = 0; i < maybeSubpath.length; i++) {
      if (maybeSubpath[i] !== path[i]) return false;
    }
    return true;
  }

  /**
   * @param {string} string
   * @param {number} length
   */
  static elide(string, length) {
    if (string.length <= length) return string;
    return string.slice(0, length - 1) + '…';
  }

  /**
   * @template {string} T
   * @param {T} name
   * @param {string=} className
   * @param {Object<string, (string|undefined)>=} attrs Attribute key/val pairs.
   *     Note: if an attribute key has an undefined value, this method does not
   *     set the attribute on the node.
   * @return {HTMLElementByTagName[T]}
   */
  static createElement(name, className, attrs = {}) {
    const element = document.createElement(name);
    if (className) {
      element.className = className;
    }
    Object.keys(attrs).forEach(key => {
      const value = attrs[key];
      if (typeof value !== 'undefined') {
        element.setAttribute(key, value);
      }
    });
    return element;
  }

  /**
   * @template {string} T
   * @param {Element} parentElem
   * @param {T} elementName
   * @param {string=} className
   * @param {Object<string, (string|undefined)>=} attrs Attribute key/val pairs.
   *     Note: if an attribute key has an undefined value, this method does not
   *     set the attribute on the node.
   * @return {HTMLElementByTagName[T]}
   */
  static createChildOf(parentElem, elementName, className, attrs) {
    const element = this.createElement(elementName, className, attrs);
    parentElem.appendChild(element);
    return element;
  }

  /**
   * Guaranteed context.querySelector. Always returns an element or throws if
   * nothing matches query.
   * @template {string} T
   * @param {T} query
   * @param {ParentNode=} context
   */
  static find(query, context = document) {
    const result = context.querySelector(query);
    if (result === null) {
      throw new Error(`query ${query} not found`);
    }
    return result;
  }

  /**
   * @param {number} bytes
   */
  static formatBytes(bytes) {
    if (bytes >= MiB) return (bytes / MiB).toFixed(2) + '\xa0MiB';
    if (bytes >= KiB) return (bytes / KiB).toFixed(0) + '\xa0KiB';
    return bytes + '\xa0B';
  }

  /**
   * @param {number} value
   * @param {string} unit
   */
  static format(value, unit) {
    if (unit === 'bytes') return TreemapUtil.formatBytes(value);
    if (unit === 'time') return `${value}\xa0ms`;
    return `${value} ${unit}`;
  }

  /**
   * Given a list of items, return a function (a hasher) that will map keys to an item.
   * When a key is seen for the first time, the item returned is cached and will always
   * be returned for the same key.
   * The hash function is stable and deterministic, so the same key->item mapping will be
   * produced given the same call order.
   * @template T
   * @param {T[]} items
   * @return {(key: string) => T|undefined}
   */
  static stableHasher(items) {
    // Clone.
    items = [...items];

    /** @type {Map<string, T>} */
    const assignedItems = new Map();
    return key => {
      // Key has already been assigned an item.
      if (assignedItems.has(key)) return assignedItems.get(key);

      // Ran out of items.
      if (items.length === 0) return;

      // Select a random item using a stable hash.
      const hash = [...key].reduce((acc, char) => acc + char.charCodeAt(0), 0);
      const [assignedItem] = items.splice(hash % items.length, 1);
      assignedItems.set(key, assignedItem);

      return assignedItem;
    };
  }

  /**
   * @param {number} h
   * @param {number} s
   * @param {number} l
   */
  static hsl(h, s, l) {
    return `hsl(${h}, ${s}%, ${l}%)`;
  }
}

// From DevTools:
// https://cs.chromium.org/chromium/src/third_party/devtools-frontend/src/front_end/quick_open/CommandMenu.js?l=255&rcl=ad5c586c30a6bc55962b7a96b0533911c86bd4fc
// https://gist.github.com/connorjclark/f114ef39fd98f8a1b89dab2bd873d2c2
TreemapUtil.COLOR_HUES = [
  4.1,
  339.6,
  291.2,
  261.6,
  230.8,
  198.7,
  186.8,
  174.4,
  122.4,
  87.8,
  65.5,
  45,
  35.8,
  14.4,
  15.9,
  0,
  199.5,
];

// node export for testing.
if (typeof module !== 'undefined' && module.exports) {
  module.exports = TreemapUtil;
}
